package nervoussystem.nurbs;

import processing.core.*;

/*
 * NurbsCurve
 * by Nervous System
 * 
 * Implementation of NURBS curves
 * Use at your own discretion
 * 
 * Current state: basic display of NURBS curves.  Interface only allows B-Splines, but implementation allows full NURBS capabilities
 * 
 * To do: derivatives, knot insertion, knot deletion, curve interpolation, degree elevation
 * 
 */


public class NurbsCurve {
	public static float resolution = (float) 0.01;
	protected PVector[] controlPts = new PVector[0];
	protected float[] knots = new float[0];
	protected float[] weights = new float[0];
	protected int degree = 3;
	protected boolean closed = false;
	private float[][] B;
	
	public NurbsCurve() {
	}
	
	public NurbsCurve(PVector[] pts) {
	  controlPts = pts;
	  degree = 3;
	  knots = new float[pts.length+degree+1];
	  weights = new float[pts.length];
	  setB();
	  for(int i=0;i<pts.length;++i) {
	    weights[i] = 1;
	  }
	  for(int i=0;i<=degree;++i) {
	    knots[i] = 0;
	  }
	  for(int i=1;i<pts.length-degree;++i) {
	    //knots[i+degree] = float(i)/(pts.length+1.0);
	    knots[i+degree] = i;
	  }
	  for(int i=PApplet.max(knots.length-(degree+1),degree+1);i<knots.length;++i) {
	    knots[i] = PApplet.max(pts.length-degree,1);
	  }
	  
	}
	
	public NurbsCurve(PVector[] pts, int deg) {
		  controlPts = pts;
		  degree = deg;
		  knots = new float[pts.length+degree+1];
		  weights = new float[pts.length];
		  setB();
		  for(int i=0;i<pts.length;++i) {
		    weights[i] = 1;
		  }
		  for(int i=0;i<=degree;++i) {
		    knots[i] = 0;
		  }
		  for(int i=1;i<pts.length-degree;++i) {
		    //knots[i+degree] = float(i)/(pts.length+1.0);
		    knots[i+degree] = i;
		  }
		  for(int i=PApplet.max(knots.length-(degree+1),degree+1);i<knots.length;++i) {
		    knots[i] = PApplet.max(pts.length-degree,1);
		  }
		}
	  
	public  NurbsCurve(PVector[] pts, float[] _knots, float[] _weights, int deg) {
	    controlPts = pts;
	    knots = _knots;
	    degree = deg;
	    weights = _weights;
	    setB();
	  }
	
	public NurbsCurve(NurbsCurve crv) {
		controlPts = new PVector[crv.controlPts.length];
		knots = new float[crv.knots.length];
		weights = new float[crv.weights.length];
		PApplet.arrayCopy(crv.controlPts,controlPts);
		PApplet.arrayCopy(crv.knots,knots);
		PApplet.arrayCopy(crv.weights,weights);
		degree = crv.degree;
		B= crv.B;
	}
	
	public static NurbsCurve circle(PVector center, float radius) {
		PVector pts[] = new PVector[] {
				PVector.add(center, new PVector(radius,0,0)),
				PVector.add(center, new PVector(radius,radius,0)),
				PVector.add(center, new PVector(0,radius,0)),
				PVector.add(center, new PVector(-radius,radius,0)),
				PVector.add(center, new PVector(-radius,0,0)),
				PVector.add(center, new PVector(-radius,-radius,0)),
				PVector.add(center, new PVector(0,-radius,0)),
				PVector.add(center, new PVector(radius,-radius,0)),
				PVector.add(center, new PVector(radius,0,0))
		};
		float  knots[] = new float[] {
			0,0,0,PApplet.HALF_PI,PApplet.HALF_PI, PApplet.PI,PApplet.PI,1.5f*PApplet.PI,1.5f*PApplet.PI,PApplet.TWO_PI,PApplet.TWO_PI,PApplet.TWO_PI
		};
		float SQRT2 = PApplet.sqrt(2.0f)/2.0f;
		float weights[] = new float[] {
			1,SQRT2,1,SQRT2,1,SQRT2,1,SQRT2,1
		};
		return new NurbsCurve(pts,knots,weights,2);
	}
	
	public PVector getPoint(int index) {
		return controlPts[index];
	}
	
	public float getWeight(int index) {
		return weights[index];
	}
	
	public int numPoints() {
		return controlPts.length;
	}
	
	public int degree() {
		return degree;
	}
	
	public void setPoint(int index,PVector pt) {
		controlPts[index] = pt;
	}
	
	public void setWeight(int index, float w) {
		weights[index] = w;
	}
	
	public void setDegree(int deg) {
		//implement degree elevation and reduction
	}
	
	 public PVector evaluate(float u) {
	    int currKnot = findKnot(u);
	    float[] basFunc = basisFunctions(currKnot, u);
	    Vector4D evalPt = new Vector4D();
	    evalPt.w = 0;
	    Vector4D[] homoCtrlPts = new Vector4D[degree+1];
	    for(int i=0;i<=degree;++i) {
	      homoCtrlPts[i] = Vector4D.multiply(new Vector4D(controlPts[currKnot-degree+i].x,controlPts[currKnot-degree+i].y,controlPts[currKnot-degree+i].z),weights[currKnot-degree+i]);
	    }
	    for(int i = 0;i<=degree;++i) {
	      evalPt = Vector4D.add(evalPt,Vector4D.multiply(homoCtrlPts[i],basFunc[i]));
	    }
	    return new PVector(evalPt.x/evalPt.w,evalPt.y/evalPt.w,evalPt.z/evalPt.w);
	  }
	 
	 public PVector derivative(float u, int k) {
		 Vector4D[] derivesW = new Vector4D[k+1];
		 if(k>degree) return new PVector();
		 int currKnot = findKnot(u);
		 Vector4D[] hPts = new Vector4D[degree+1];
		 for(int i=0;i<=degree;++i) {
	      hPts[i] = Vector4D.multiply(new Vector4D(controlPts[currKnot-degree+i].x,controlPts[currKnot-degree+i].y,controlPts[currKnot-degree+i].z),weights[currKnot-degree+i]);
		 }		 
		 float[][] basFunc = deriveBasisFunctions(currKnot,u, k);
		 for(int i=0;i<=k;++i) {
			 derivesW[i] = new Vector4D();
			 for(int j=0;j<=degree;++j) {
				 derivesW[i] = Vector4D.add(derivesW[i],Vector4D.multiply(hPts[j],basFunc[i][j]));
			 }
		 }
		 
		 PVector[] derives = new PVector[derivesW.length];
		 derives[0] = new PVector();
		 for(int i=0;i<=k;++i) {
			PVector currPt = new PVector(derivesW[i].x,derivesW[i].y,derivesW[i].z);
			for(int j=1;j<=i;++j) {
				currPt = PVector.sub(currPt,PVector.mult(derives[i-j],B[i][j]*derivesW[j].w));
			}
			derives[i] = new PVector(currPt.x/derivesW[0].w,currPt.y/derivesW[0].w,currPt.z/derivesW[0].w);
		 }
		 return derives[k];
		 
	 }
	 
	 public PVector[] allDerivatives(float u, int k) {
		 Vector4D[] derivesW = new Vector4D[k+1];
		 int currKnot = findKnot(u);
		 Vector4D[] hPts = new Vector4D[degree+1];
		 for(int i=0;i<=degree;++i) {
	      hPts[i] = Vector4D.multiply(new Vector4D(controlPts[currKnot-degree+i].x,controlPts[currKnot-degree+i].y,controlPts[currKnot-degree+i].z),weights[currKnot-degree+i]);
		 }		 
		 float[][] basFunc = deriveBasisFunctions(currKnot,u, k);
		 for(int i=0;i<=k;++i) {
			 derivesW[i] = new Vector4D();
			 for(int j=0;j<=degree;++j)
				 derivesW[i] = Vector4D.add(derivesW[i],Vector4D.multiply(hPts[j],basFunc[i][j]));
		 }
		 
		 PVector[] derives = new PVector[derivesW.length];
		 derives[0] = new PVector();
		 for(int i=0;i<=k;++i) {
			PVector currPt = new PVector(derivesW[i].x,derivesW[i].y,derivesW[i].z);
			for(int j=1;j<=i;++j) {
				currPt = PVector.sub(currPt,PVector.mult(derives[i-j],B[i][j]*derivesW[j].w));
			}
			derives[i] = new PVector(currPt.x/derivesW[0].w,currPt.y/derivesW[0].w,currPt.z/derivesW[0].w);
		 }
		 return derives;
		 
	 }	 
	  public void draw2D(PGraphics g, int divs) {
	    float start = knots[0];
	    float end = knots[knots.length-1];
	    float inc = (end-start)/divs;
	    g.beginShape();
	    
	    for(int i=0;i<=divs;++i) {
	      PVector pt = evaluate(i*inc+start);
	      g.vertex(pt.x,pt.y);
	    }
	    g.endShape();
	  }
		  
	  public void draw2D(PGraphics g) {
		  draw2D(g,20);
	  }
	  
	  public void draw(PGraphics g, int divs) {
		    float start = knots[0];
		    float end = knots[knots.length-1];
		    float inc = (end-start)/divs;
		    g.beginShape();
		    for(int i=0;i<=divs;++i) {
		      PVector pt = evaluate(i*inc+start);
		      g.vertex(pt.x,pt.y,pt.z);
		    }
		    g.endShape();		  
	  }
	  
	  public void draw(PGraphics g) {
		  draw(g,20);
	  }
		  
	  public float length() {
	    return 1;
	  }	
	  
	  public float[] domain() {
		  return new float[] {knots[0],knots[knots.length-1]};
	  }
	  
	  public void addPoint(PVector pt) {
		  controlPts = (PVector[]) PApplet.append(controlPts, pt);
		  weights = PApplet.append(weights, 1);
		  float inc = 1;
		  float start = knots[degree];
		  float end = knots[knots.length-1];
		  if(controlPts.length<=degree+1) {
              knots = PApplet.append(knots,1);
          } else {		  
			  for(int i=degree+1;i<knots.length-degree;++i) {
				  if(knots[i] != start) {
					  inc = knots[i]-start;
					  i= knots.length;
				  }
			  }
			  knots = PApplet.append(knots, end+inc);
			  for(int i=knots.length-2;i>knots.length-degree-2;--i) 
				knots[i] = end+inc;			  
			  for(int i=0;i<knots.length;++i) 
				knots[i] /= end+inc;
          }
	 }
	  
	 public void insertKnot(float u) {
		if(u<= knots[0] || u >= knots[knots.length-1] ) return;
	 	int currKnot = findKnot(u);
	 	//float[] knots2 = new float[knots.length+1];
	 	PVector[] newPt = new PVector[controlPts.length+1];
	 	float[] newWeight = new float[weights.length+1];
	 	for(int i=0;i<currKnot-degree+1;++i) {
	 		newPt[i] = controlPts[i];
	 		newWeight[i] = weights[i];
	 	}
	 	for(int i=currKnot-degree+1;i<=currKnot;++i) {
	 		float alpha = (u-knots[i])/(knots[i+degree]-knots[i]);
	 		newPt[i] = PVector.add(PVector.mult(controlPts[i],alpha), PVector.mult(controlPts[i-1],1-alpha));
	 		newWeight[i] = alpha*weights[i]+(1-alpha)*weights[i-1];
	 	}
	 	for(int i=currKnot;i<controlPts.length;++i) {
	 		newPt[i+1] = controlPts[i];
	 		newWeight[i+1] = weights[i];
	 		
	 	}
	 	/*controlPts = (PVector[]) PApplet.append(controlPts, controlPts[controlPts.length-1]);
	 	weights = PApplet.append(weights, weights[weights.length-1]);
	 	for(int i=controlPts.length-1;i>currKnot;--i) {
	 		controlPts[i] = controlPts[i-1];
	 		weights[i] = weights[i-1];
	 	}
	 	for(int i=currKnot;i>=currKnot-degree+1;--i) {
	 		controlPts[i] = newPt[i-(currKnot-degree+1)];
	 		weights[i] = newWeight[i-(currKnot-degree+1)];
	 	}
	 	*/
	 	controlPts = newPt;
	 	weights = newWeight;
	 	knots = PApplet.splice(knots, u, currKnot+1);
	 }
	 
	 public void insertKnot(float u, int times) {
		 int currKnot = findKnot(u);
		 int multiplicity = findMultiplicity(currKnot);
		 //times = Math.min(degree-times-multiplicity,times);
		 //times = Math.max(0,times);
		 float[] newKnots = new float[knots.length+times];
		 Vector4D[] newPoints = new Vector4D[controlPts.length+times];
		 
		 for(int i=0;i<=currKnot;++i) newKnots[i] = knots[i];
		 for(int i=1;i<=times;++i) newKnots[currKnot+i] = u;
		 for(int i=currKnot+1;i<knots.length;++i) newKnots[i+times] = knots[i];
		 for(int i=0;i<=currKnot-degree;++i) newPoints[i] = new Vector4D(controlPts[i],weights[i]);
		 for(int i=currKnot-multiplicity; i<controlPts.length;++i) newPoints[i+times] = new Vector4D(controlPts[i],weights[i]);
		 Vector4D[] temp = new Vector4D[degree+1];
		 for(int i=0;i<= degree-multiplicity;++i) temp[i] = new Vector4D(controlPts[currKnot-degree+i],weights[currKnot-degree+i]);
		 int L = 0;
		 for(int j=1;j<=times;++j) {
			 L = currKnot-degree+j;
		 	/*for(int i=L;i<=currKnot-multiplicity;++i) {
		 		float alpha = (u-knots[i])/(knots[i+degree]-knots[i]);
		 		temp[i-L] =Vector4D.add(Vector4D.multiply(temp[i-L+1],alpha), Vector4D.multiply(temp[i-L],1-alpha));
		 		
		 	}
		 	*/
			 for(int i=0;i<=degree-j-multiplicity;++i) {
				 float alpha = (u-knots[L+i])/(knots[i+currKnot+1]-knots[L+i]);
				 temp[i] = Vector4D.add(Vector4D.multiply(temp[i+1],alpha),Vector4D.multiply(temp[i],1-alpha));
			 }
			 
			 newPoints[L] = temp[0];
			 newPoints[currKnot+times-j-multiplicity] = temp[degree-j-multiplicity];
		 }
		 for(int i=L+1;i<currKnot-multiplicity;++i) {
			 newPoints[i] = temp[i-L];
		 }
		 controlPts = new PVector[newPoints.length];
		 weights = new float[newPoints.length];
		 for(int i=0;i<newPoints.length;++i) {
			 
			 if(newPoints[i] != null) {
				 controlPts[i] = newPoints[i].projectDown();
				 weights[i] = newPoints[i].w;
			 } else {
			 }
		 }
		 knots = newKnots;
	 }
	 
	 public void insertKnots(float[] insertKnots) {
		 int startKnot = findKnot(insertKnots[0]);
		 int endKnot = findKnot(insertKnots[insertKnots.length-1])+1;
		 float[] newKnots = new float[knots.length+insertKnots.length];
		 Vector4D[] newPoints = new Vector4D[controlPts.length+insertKnots.length];
		 for(int j=0;j<=startKnot-degree;++j) newPoints[j] = new Vector4D(controlPts[j],weights[j]);
		 for(int j=endKnot-1;j<controlPts.length;++j) newPoints[j+insertKnots.length] =  new Vector4D(controlPts[j],weights[j]);
		 for(int j=0;j<=startKnot;++j) newKnots[j] = knots[j];
		 for(int j=endKnot+degree;j<knots.length;++j) newKnots[j+insertKnots.length] = knots[j];
		 int i=endKnot+degree-1;
		 int k= endKnot+degree+insertKnots.length-1;
		 for(int j=insertKnots.length-1;j>=0;--j) {
			 while(insertKnots[j] <= knots[i] && i>startKnot) {
				 newPoints[k-degree-1] = new Vector4D(controlPts[i-degree-1],weights[i-degree-1]);
				 newKnots[k] = knots[i];
				 --k;
				 --i;
			 }
			 newPoints[k-degree-1] = newPoints[k-degree];
			 for(int l=1;l<=degree;++l) {
				 int ind = k-degree+l;
				 float alpha = newKnots[k+l]-insertKnots[j];
				 if(Math.abs(alpha) == 0) newPoints[ind-1] = newPoints[ind];
				 else {
					 alpha = alpha/(newKnots[k+l]-knots[i-degree+l]);
					 newPoints[ind-1] = Vector4D.add(Vector4D.multiply(newPoints[ind-1],alpha), Vector4D.multiply(newPoints[ind],1-alpha));
				 }
			 }
			 newKnots[k] = insertKnots[j];
			 --k;
		 }
		 knots = newKnots;
		 controlPts = new PVector[newPoints.length];
		 weights = new float[newPoints.length];
		 for(int j=0;j<newPoints.length;++j) {
			 
			 if(newPoints[j] != null) {
				 controlPts[j] = newPoints[j].projectDown();
				 weights[j] = newPoints[j].w;
			 }
		 }
	 }
	 
	 protected void normalizeKnots() {
		 float start = knots[0];
		 float end = knots[knots.length-1];
		 for(int i=0;i<knots.length;++i) {
		 	 knots[i] = (knots[i]-start)/(end-start);
		 }
	 }
	 
	 public int findMultiplicity(int knot) {
		int mult = 1;
		for(int i=knot+1;i<knots.length && knots[i] == knots[knot];++i) ++mult;
		for(int i=knot-1;i>=0 && knots[i] == knots[knot];--i) ++mult;

		return mult-1;
	 }
	 
	 private float[] basisFunctions( int knot, float u ) {
	    float[] funcs = new float[degree+1];
	    float[] left = new float[degree+1];
	    float[] right = new float[degree+1];

	    funcs[0] = 1;
	    for(int j=1;j<=degree;++j) {
	      left[j] = u-knots[knot+1-j];
	      right[j] = knots[knot+j]-u;
	      float saved = 0;
	      for(int r = 0;r<j;++r) {
	        float temp = funcs[r]/(right[r+1]+left[j-r]);
	        funcs[r] = saved+right[r+1]*temp;
	        saved = left[j-r]*temp;
	      }
	      funcs[j] = saved;
	    }
	    return funcs;
	  }
	  
	 private float[][] deriveBasisFunctions(int knot, float u, int der) {
		 float[][] ndu = new float[degree+1][degree+1];
		 float[] left = new float[degree+1];
		 float[] right = new float[degree+1];
		 float[][] ders = new float[degree+1][degree+1];
		 ndu[0][0] = 1;
		 for(int j=1;j<=degree;++j) {
			 left[j] = u-knots[knot+1-j];
			 right[j] = knots[knot+j]-u;
			 float saved = 0;
			 for(int r=0;r<j;++r) {
				 ndu[j][r] = right[r+1]+left[j-r];
				 float temp = ndu[r][j-1]/ndu[j][r];
				 ndu[r][j] = saved+right[r+1]*temp;
				 saved = left[j-r]*temp;
			 }
			 ndu[j][j] = saved;
		 }
		 for(int j=0;j<=degree;++j)
			 	ders[0][j] = ndu[j][degree];
		 for(int r=0;r<=degree;++r) {
			 int s1 = 0;
			 int s2 = 1;
			 float[][] a = new float[degree+1][degree+1];
			 a[0][0] = 1;
			 for(int k=1;k<=der;++k) {
				 float d = 0;
				 int rk = r-k;
				 int pk = degree-k;
				 if(r>=k) {
					 a[s2][0] = a[s1][0]/ndu[pk+1][rk];
					 d = a[s2][0]*ndu[rk][pk];
				 }
				 int j1 = -rk;
				 if(rk>=-1) j1 = 1;
				 int j2=degree-r;
				 if(r-1 <=pk) j2 = k-1;
				 
				 for(int j=j1;j<=j2;++j) {
					 a[s2][j] = (a[s1][j]-a[s1][j-1])/ndu[pk+1][rk+j];
					 d += a[s2][j]*ndu[rk+j][pk];
				 }
				 if(r<=pk) {
					 a[s2][k] = -a[s1][k-1]/ndu[pk+1][r];
					 d += a[s2][k]*ndu[r][pk];
				 }
				 ders[k][r] = d;
				 int temp =s1;
				 s1 = s2;
				 s2 = temp;	 
			 }
		 }
		 int r = degree;
		 for(int k=1;k<=der;++k) {
			 for(int j=0;j<=degree;++j) ders[k][j] *= r; 
			 r *= (degree-k);
		 }
		 return ders;
	 }
	 
	 public int findKnot(float u) {
	    if (u==knots[knots.length-degree]) return knots.length-degree-2;
	    if(u <= knots[degree]) return degree;
	    int low = degree;
	    int high = knots.length-degree;
	    int mid = (high+low)/2;
	    while(knots[mid]>u || u >= knots[mid+1]) {
		  if(u<knots[mid]) {
	        high = mid;
	      } else {
	        low = mid;
	      }
	      mid = (high+low)/2;
	    }
	    return mid;
	  }
	 
	 private void setB() {
		 B = new float[10][10];
		 for(int i=0;i<10;++i) {
			 for(int j=0;j<10;++j) {
				 B[i][j] = bang(i)/(bang(j)*bang(i-j));
			 }
		 }
	 }
	 
	 private float bang(int val) {
		 if(val <= 1) return 1;
		 return val*bang(val-1);
	 }
}
