package nervoussystem.nurbs;

import processing.core.*;
/*
 * NurbsCurve
 * by Nervous System
 * 
 * Implementation of NURBS surfaces
 * Use at your own discretion
 * 
 * Current state: nothing
 * 
 * To do: everything, evaluate, derivatives, sweeps, revolutions, interpolations
 * 
 */

public class NurbsSurface {
	public static float resolution = 20;
	protected float knotsU[];
	protected float knotsV[];
	protected float weights[][];
	protected PVector controlPts[][];
	protected int degreeU;
	protected int degreeV;
	
	public NurbsSurface(PVector[][] pts) {
		this(pts,3,3);
	}
	
	public NurbsSurface(PVector[][] pts, int degu, int degv) {
		controlPts = pts;
		degreeU = degu;
		degreeV = degv;
		int nu = pts.length;
		int nv = 0;
		if(nu>0) nv = pts[0].length;
		weights = new float[nu][nv];
		knotsU = new float[nu+degreeU+1];
		knotsV = new float[nv+degreeV+1];
		for(int i=0;i<nu;++i) {
			for(int j=0;j<nv;++j) {
				weights[i][j] = 1;
			}
		}
		for(int i=0;i<=degreeU;++i) {
			  knotsU[i] = 0;
		}
		for(int i=1;i<nu-degreeU;++i) {
		  //knots[i+degree] = float(i)/(pts.length+1.0);
		  knotsU[i+degreeU] = i;
		}
		for(int i=PApplet.max(knotsU.length-(degreeU+1),degreeU+1);i<knotsU.length;++i) {
		  knotsU[i] = PApplet.max(nu-degreeU,1);
		}
		for(int i=0;i<degreeV;++i) {
		   knotsV[i] = 0;
		}
		for(int i=1;i<nv-degreeV;++i) {
		  //knots[i+degree] = float(i)/(pts.length+1.0);
		  knotsV[i+degreeV] = i;
		}
		for(int i=PApplet.max(knotsV.length-(degreeV+1),degreeV+1);i<knotsV.length;++i) {
		  knotsV[i] = PApplet.max(nv-degreeV,1);
		}
	}
	
	public float[][] domain() {
		return new float[][] {new float[] {knotsU[0], knotsU[knotsU.length-1]}, new float[] {knotsV[0],knotsV[knotsV.length-1]}};
	}
	
	public PVector getPoint(int indU, int indV) {
		return controlPts[indU][indV];
	}
	
	public float getWeight(int indU, int indV) {
		return weights[indU][indV];
	}
	
	public int numPointsU() {
		return controlPts.length;
	}
	
	public int numPointsV() {
		if(controlPts.length == 0) return 0;
		return controlPts[0].length;
	}
	
	public void setPoint(int indU, int indV, PVector pt) {
		controlPts[indU][indV] = pt;
	}
	
	public void setWeight(int indU, int indV, float w){
		weights[indU][indV] = w;
	}
	
	PVector evaluate(float u, float v) {
		if(controlPts.length == 0) return new PVector();
		int uKnot = findKnot(u,knotsU,degreeU);
		int vKnot = findKnot(v,knotsV,degreeV);
		float[] basFuncU = basisFunctions(uKnot,u,knotsU,degreeU);
		float[] basFuncV = basisFunctions(vKnot,v,knotsV,degreeV);
		Vector4D[][] hPts = new Vector4D[degreeU+1][degreeV+1];
		for(int i=0;i<=degreeV;++i) {
			for(int j=0;j<=degreeU;++j) {
				PVector ctrlPt = controlPts[uKnot-degreeU+j][vKnot-degreeV+i];
				float w = weights[uKnot-degreeU+j][vKnot-degreeV+i];
				hPts[j][i] = new Vector4D(ctrlPt.x*w, ctrlPt.y*w,ctrlPt.z*w,w);
			}
		}

		Vector4D[] temp = new Vector4D[degreeU+1];
		for(int i=0;i<=degreeV;++i) {
			temp[i] = new Vector4D();
			for(int j=0;j<=degreeU;++j) {
				temp[i] = Vector4D.add(temp[i],Vector4D.multiply(hPts[j][i],basFuncU[j]));
			}
		}
		
		Vector4D evalPt = new Vector4D();
		for(int i=0;i<=degreeV;++i) {
			evalPt = Vector4D.add(evalPt, Vector4D.multiply(temp[i],basFuncV[i]));
		}
		return new PVector(evalPt.x/evalPt.w,evalPt.y/evalPt.w,evalPt.z/evalPt.w);
	}
	
	//ADD OPTION FOR V or U
	NurbsCurve isocurve(float u, boolean dir) {
		int uKnot = findKnot(u,knotsU,degreeU);
		float[] basFunc = basisFunctions(uKnot,u,knotsU,degreeU);
		Vector4D[][] hPts = new Vector4D[degreeU+1][degreeV+1];
		for(int i=0;i<controlPts.length;++i) {
			for(int j=0;j<controlPts[0].length;++j) {
				PVector ctrlPt = controlPts[i][j];
				float w = weights[i][j];
				hPts[i][j] = new Vector4D(ctrlPt.x*w, ctrlPt.y*w,ctrlPt.z*w,w);
			}
		}
		Vector4D[] newPts = new Vector4D[controlPts[0].length];
		for(int i=0;i<controlPts[0].length;++i) {
			for(int j=0;j<=degreeU;++j) {
				newPts[i] = Vector4D.add(newPts[i],Vector4D.multiply(hPts[uKnot-degreeU+j][i], basFunc[j]));
			}
		}
		
		PVector[] newCPts = new PVector[newPts.length];
		float[] newWeights = new float[newPts.length];
		for(int i=0;i<newPts.length;++i) {
			newCPts[i] = new PVector(newPts[i].x*newPts[i].w,newPts[i].y*newPts[i].w,newPts[i].z*newPts[i].w);
			newWeights[i] = newPts[i].w;
		}
		return new NurbsCurve(newCPts, knotsV, newWeights, degreeV);
	}
	
	public void draw(PGraphics g) {
		float[][] dom = domain();
		PVector[][] evalPts = new PVector[(int)resolution][(int)resolution];
		for(int i=0;i<evalPts.length;++i) {
			for(int j=0;j<evalPts[i].length;++j) {
				evalPts[i][j] = evaluate(((float)i/(float)(evalPts.length-1))*(dom[0][1]-dom[0][0])+dom[0][0],((float)j/(float)(evalPts[i].length-1))*(dom[1][1]-dom[1][0])+dom[1][0]);
			}
		}
		for(int i=0;i<evalPts.length-1;++i) {
			for(int j=0;j<evalPts[i].length-1;++j) {
				PVector pt1 = evalPts[i][j];
				PVector pt2 = evalPts[i+1][j];
				PVector pt3 = evalPts[i+1][j+1];
				PVector pt4 = evalPts[i][j+1];
				g.beginShape();
				g.vertex(pt1.x,pt1.y,pt1.z);
				g.vertex(pt2.x,pt2.y,pt2.z);
				g.vertex(pt4.x,pt4.y,pt4.z);
				g.vertex(pt1.x,pt1.y,pt1.z);
				g.endShape();
				g.beginShape();
				g.vertex(pt3.x,pt3.y,pt3.z);
				g.vertex(pt4.x,pt4.y,pt4.z);
				g.vertex(pt2.x,pt2.y,pt2.z);
				g.vertex(pt3.x,pt3.y,pt3.z);
				g.endShape();
			}
		}
	}
	
	private int findKnot(float u, float[] knots, int degree) {
	    if (u==knots[knots.length-degree]) return knots.length-degree-2;
	    int low = degree;
	    int high = knots.length-degree;
	    int mid = (high+low)/2;
	    while(knots[mid]>u || u >= knots[mid+1]) {
	      if(u<knots[mid]) {
	        high = mid;
	      } else {
	        low = mid;
	      }
	      mid = (high+low)/2;
	    }
	    return mid;
	}
	
	private float[] basisFunctions( int knot, float u, float knots[], int degree ) {
	    float[] funcs = new float[degree+1];
	    float[] left = new float[degree+1];
	    float[] right = new float[degree+1];

	    funcs[0] = 1;
	    for(int j=1;j<=degree;++j) {
	      left[j] = u-knots[knot+1-j];
	      right[j] = knots[knot+j]-u;
	      float saved = 0;
	      for(int r = 0;r<j;++r) {
	        float temp = funcs[r]/(right[r+1]+left[j-r]);
	        funcs[r] = saved+right[r+1]*temp;
	        saved = left[j-r]*temp;
	      }
	      funcs[j] = saved;
	    }
	    return funcs;
	}
	
	 /*
	  *  static surface construction
	  */
	
	public static NurbsSurface loft(NurbsCurve crv1, NurbsCurve crv2) {
		if(crv1.degree != crv2.degree) return null;
		NurbsCurve temp1 = new NurbsCurve(crv1);
		NurbsCurve temp2 = new NurbsCurve(crv2);
		temp1.normalizeKnots();
		temp2.normalizeKnots();
		//find difference
		int k = 0;
		float[] insertTemp1 = new float[0];
		float[] insertTemp2 = new float[0];
		for(int i=0;i<temp1.knots.length;++i) {
			while(k < temp2.knots.length &&temp2.knots[k] < temp1.knots[i] ) {
				insertTemp1 = PApplet.append(insertTemp1, temp2.knots[k]);
				++k;
			}
			if(temp2.knots[k] > temp1.knots[i]) insertTemp2 = PApplet.append(insertTemp2,temp1.knots[i]);
			if(temp2.knots[k] == temp1.knots[i]) ++k;
		}
		while(k<temp2.knots.length) {
			insertTemp1 = PApplet.append(insertTemp1,temp2.knots[k]);
			++k;
		}
		
		if(insertTemp1.length > 0) temp1.insertKnots(insertTemp1);
		if(insertTemp2.length > 0) temp2.insertKnots(insertTemp2);
		PVector[][] pts = new PVector[temp1.controlPts.length][2];
		for(int i=0;i<pts.length;++i) {
			pts[i][0] = temp1.controlPts[i];
			pts[i][1] = temp2.controlPts[i];
		}
		NurbsSurface toReturn = new NurbsSurface(pts,temp1.degree,1);
		for(int i=0;i<temp1.knots.length;++i) {
			toReturn.knotsU[i] = temp1.knots[i];
		}
		for(int i=0;i<pts.length;++i) {
			toReturn.weights[i][0] = temp1.weights[i];
			toReturn.weights[i][1] = temp2.weights[i];
		}
		return toReturn;
	}
}
